@@ -18,8 +18,8 @@ import static org.assertj.core.api.Assertions.assertThat;
 public class ColumnFamilyTest {
 
   @ClassRule
-  public static final RocksMemoryResource rocksMemoryResource =
-      new RocksMemoryResource();
+  public static final RocksNativeLibraryResource ROCKS_NATIVE_LIBRARY_RESOURCE =
+      new RocksNativeLibraryResource();
 
   @Rule
   public TemporaryFolder dbFolder = new TemporaryFolder();
@@ -75,6 +75,7 @@ public class ColumnFamilyTest {
 
         assertThat(cfh.getName()).isEqualTo("default".getBytes(UTF_8));
         assertThat(cfh.getID()).isEqualTo(0);
+        assertThat(cfh.getDescriptor().getName()).isEqualTo("default".getBytes(UTF_8));
 
         final byte[] key = "key".getBytes();
         final byte[] value = "value".getBytes();
@@ -154,10 +155,10 @@ public class ColumnFamilyTest {
         assertThat(retVal).isEqualTo("newcfvalue");
         assertThat((db.get(columnFamilyHandleList.get(1),
             "dfkey1".getBytes()))).isNull();
-        db.remove(columnFamilyHandleList.get(1), "newcfkey1".getBytes());
+        db.delete(columnFamilyHandleList.get(1), "newcfkey1".getBytes());
         assertThat((db.get(columnFamilyHandleList.get(1),
             "newcfkey1".getBytes()))).isNull();
-        db.remove(columnFamilyHandleList.get(0), new WriteOptions(),
+        db.delete(columnFamilyHandleList.get(0), new WriteOptions(),
             "dfkey2".getBytes());
         assertThat(db.get(columnFamilyHandleList.get(0), new ReadOptions(),
             "dfkey2".getBytes())).isNull();
@@ -229,6 +230,7 @@ public class ColumnFamilyTest {
                 new ColumnFamilyOptions()));
         db.put(tmpColumnFamilyHandle, "key".getBytes(), "value".getBytes());
         db.dropColumnFamily(tmpColumnFamilyHandle);
+        assertThat(tmpColumnFamilyHandle.isOwningHandle()).isTrue();
       } finally {
         if (tmpColumnFamilyHandle != null) {
           tmpColumnFamilyHandle.close();
@@ -240,6 +242,46 @@ public class ColumnFamilyTest {
     }
   }
 
+  @Test
+  public void createWriteDropColumnFamilies() throws RocksDBException {
+    final List<ColumnFamilyDescriptor> cfDescriptors = Arrays.asList(
+        new ColumnFamilyDescriptor(RocksDB.DEFAULT_COLUMN_FAMILY),
+        new ColumnFamilyDescriptor("new_cf".getBytes()));
+    final List<ColumnFamilyHandle> columnFamilyHandleList = new ArrayList<>();
+    try (final DBOptions options = new DBOptions()
+        .setCreateIfMissing(true)
+        .setCreateMissingColumnFamilies(true);
+         final RocksDB db = RocksDB.open(options,
+             dbFolder.getRoot().getAbsolutePath(), cfDescriptors,
+             columnFamilyHandleList)) {
+      ColumnFamilyHandle tmpColumnFamilyHandle = null;
+      ColumnFamilyHandle tmpColumnFamilyHandle2 = null;
+      try {
+        tmpColumnFamilyHandle = db.createColumnFamily(
+            new ColumnFamilyDescriptor("tmpCF".getBytes(),
+                new ColumnFamilyOptions()));
+        tmpColumnFamilyHandle2 = db.createColumnFamily(
+            new ColumnFamilyDescriptor("tmpCF2".getBytes(),
+                new ColumnFamilyOptions()));
+        db.put(tmpColumnFamilyHandle, "key".getBytes(), "value".getBytes());
+        db.put(tmpColumnFamilyHandle2, "key".getBytes(), "value".getBytes());
+        db.dropColumnFamilies(Arrays.asList(tmpColumnFamilyHandle, tmpColumnFamilyHandle2));
+        assertThat(tmpColumnFamilyHandle.isOwningHandle()).isTrue();
+        assertThat(tmpColumnFamilyHandle2.isOwningHandle()).isTrue();
+      } finally {
+        if (tmpColumnFamilyHandle != null) {
+          tmpColumnFamilyHandle.close();
+        }
+        if (tmpColumnFamilyHandle2 != null) {
+          tmpColumnFamilyHandle2.close();
+        }
+        for (ColumnFamilyHandle columnFamilyHandle : columnFamilyHandleList) {
+          columnFamilyHandle.close();
+        }
+      }
+    }
+  }
+
   @Test
   public void writeBatch() throws RocksDBException {
     try (final StringAppendOperator stringAppendOperator = new StringAppendOperator();
@@ -268,8 +310,8 @@ public class ColumnFamilyTest {
               "value".getBytes());
           writeBatch.put(columnFamilyHandleList.get(1), "newcfkey2".getBytes(),
               "value2".getBytes());
-          writeBatch.remove("xyz".getBytes());
-          writeBatch.remove(columnFamilyHandleList.get(1), "xyz".getBytes());
+          writeBatch.delete("xyz".getBytes());
+          writeBatch.delete(columnFamilyHandleList.get(1), "xyz".getBytes());
           db.write(writeOpt, writeBatch);
 
           assertThat(db.get(columnFamilyHandleList.get(1),
@@ -355,19 +397,63 @@ public class ColumnFamilyTest {
         final List<byte[]> keys = Arrays.asList(new byte[][]{
             "key".getBytes(), "newcfkey".getBytes()
         });
-        Map<byte[], byte[]> retValues = db.multiGet(columnFamilyHandleList,
+
+        List<byte[]> retValues = db.multiGetAsList(columnFamilyHandleList, keys);
+        assertThat(retValues.size()).isEqualTo(2);
+        assertThat(new String(retValues.get(0)))
+            .isEqualTo("value");
+        assertThat(new String(retValues.get(1)))
+            .isEqualTo("value");
+        retValues = db.multiGetAsList(new ReadOptions(), columnFamilyHandleList,
             keys);
         assertThat(retValues.size()).isEqualTo(2);
-        assertThat(new String(retValues.get(keys.get(0))))
+        assertThat(new String(retValues.get(0)))
             .isEqualTo("value");
-        assertThat(new String(retValues.get(keys.get(1))))
+        assertThat(new String(retValues.get(1)))
             .isEqualTo("value");
-        retValues = db.multiGet(new ReadOptions(), columnFamilyHandleList,
+      } finally {
+        for (final ColumnFamilyHandle columnFamilyHandle :
+            columnFamilyHandleList) {
+          columnFamilyHandle.close();
+        }
+      }
+    }
+  }
+
+  @Test
+  public void multiGetAsList() throws RocksDBException {
+    final List<ColumnFamilyDescriptor> cfDescriptors = Arrays.asList(
+        new ColumnFamilyDescriptor(RocksDB.DEFAULT_COLUMN_FAMILY),
+        new ColumnFamilyDescriptor("new_cf".getBytes()));
+    final List<ColumnFamilyHandle> columnFamilyHandleList = new ArrayList<>();
+    try (final DBOptions options = new DBOptions()
+        .setCreateIfMissing(true)
+        .setCreateMissingColumnFamilies(true);
+         final RocksDB db = RocksDB.open(options,
+             dbFolder.getRoot().getAbsolutePath(),
+             cfDescriptors, columnFamilyHandleList)) {
+      try {
+        db.put(columnFamilyHandleList.get(0), "key".getBytes(),
+            "value".getBytes());
+        db.put(columnFamilyHandleList.get(1), "newcfkey".getBytes(),
+            "value".getBytes());
+
+        final List<byte[]> keys = Arrays.asList(new byte[][]{
+            "key".getBytes(), "newcfkey".getBytes()
+        });
+        List<byte[]> retValues = db.multiGetAsList(columnFamilyHandleList,
             keys);
         assertThat(retValues.size()).isEqualTo(2);
-        assertThat(new String(retValues.get(keys.get(0))))
+        assertThat(new String(retValues.get(0)))
             .isEqualTo("value");
-        assertThat(new String(retValues.get(keys.get(1))))
+        assertThat(new String(retValues.get(1)))
+            .isEqualTo("value");
+        retValues = db.multiGetAsList(new ReadOptions(), columnFamilyHandleList,
+            keys);
+        assertThat(retValues.size()).isEqualTo(2);
+        assertThat(new String(retValues.get(0)))
+            .isEqualTo("value");
+        assertThat(new String(retValues.get(1)))
             .isEqualTo("value");
       } finally {
         for (final ColumnFamilyHandle columnFamilyHandle :
@@ -505,7 +591,7 @@ public class ColumnFamilyTest {
              cfDescriptors, columnFamilyHandleList)) {
       try {
         db.dropColumnFamily(columnFamilyHandleList.get(1));
-        db.remove(columnFamilyHandleList.get(1), "key".getBytes());
+        db.delete(columnFamilyHandleList.get(1), "key".getBytes());
       } finally {
         for (final ColumnFamilyHandle columnFamilyHandle :
             columnFamilyHandleList) {
@@ -554,7 +640,7 @@ public class ColumnFamilyTest {
         keys.add("key".getBytes());
         keys.add("newcfkey".getBytes());
         final List<ColumnFamilyHandle> cfCustomList = new ArrayList<>();
-        db.multiGet(cfCustomList, keys);
+        db.multiGetAsList(cfCustomList, keys);
 
       } finally {
         for (final ColumnFamilyHandle columnFamilyHandle :
