@@ -4,7 +4,10 @@
 //  (found in the LICENSE.Apache file in the root directory).
 //
 #pragma once
-namespace rocksdb {
+
+#include "rocksdb/rocksdb_namespace.h"
+
+namespace ROCKSDB_NAMESPACE {
 namespace port {
 
 // Install a signal handler to print callstack on the following signals:
@@ -15,5 +18,11 @@ void InstallStackTraceHandler();
 // Prints stack, skips skip_first_frames frames
 void PrintStack(int first_frames_to_skip = 0);
 
+// Prints the given callstack
+void PrintAndFreeStack(void* callstack, int num_frames);
+
+// Save the current callstack
+void* SaveStack(int* num_frame, int first_frames_to_skip = 0);
+
 }  // namespace port
-}  // namespace rocksdb
+}  // namespace ROCKSDB_NAMESPACE
