@@ -1,3 +1,9 @@
+//  Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
+//  This source code is licensed under both the GPLv2 (found in the
+//  COPYING file in the root directory) and Apache 2.0 License
+//  (found in the LICENSE.Apache file in the root directory).
+//
+
 /**
  * An persistent map : key -> (list of strings), using rocksdb merge.
  * This file is a test-harness / use-case for the StringAppendOperator.
@@ -6,24 +12,27 @@
  * Copyright 2013 Facebook, Inc.
 */
 
+#include "utilities/merge_operators/string_append/stringappend.h"
+
 #include <iostream>
 #include <map>
+#include <tuple>
 
+#include "port/stack_trace.h"
 #include "rocksdb/db.h"
 #include "rocksdb/merge_operator.h"
 #include "rocksdb/utilities/db_ttl.h"
-#include "utilities/merge_operators.h"
-#include "utilities/merge_operators/string_append/stringappend.h"
-#include "utilities/merge_operators/string_append/stringappend2.h"
-#include "util/testharness.h"
+#include "test_util/testharness.h"
 #include "util/random.h"
+#include "utilities/merge_operators.h"
+#include "utilities/merge_operators/string_append/stringappend2.h"
 
-using namespace rocksdb;
+using namespace ROCKSDB_NAMESPACE;
 
-namespace rocksdb {
+namespace ROCKSDB_NAMESPACE {
 
 // Path to the database on file system
-const std::string kDbName = test::TmpDir() + "/stringappend_test";
+const std::string kDbName = test::PerThreadDBPath("stringappend_test");
 
 namespace {
 // OpenDb opens a (possibly new) rocksdb database with a StringAppendOperator
@@ -108,12 +117,26 @@ class StringLists {
 
 
 // The class for unit-testing
-class StringAppendOperatorTest : public testing::Test {
+class StringAppendOperatorTest : public testing::Test,
+                                 public ::testing::WithParamInterface<bool> {
  public:
   StringAppendOperatorTest() {
     DestroyDB(kDbName, Options());    // Start each test with a fresh DB
   }
 
+  void SetUp() override {
+#ifndef ROCKSDB_LITE  // TtlDb is not supported in Lite
+    bool if_use_ttl = GetParam();
+    if (if_use_ttl) {
+      fprintf(stderr, "Running tests with ttl db and generic operator.\n");
+      StringAppendOperatorTest::SetOpenDbFunction(&OpenTtlDb);
+      return;
+    }
+#endif  // !ROCKSDB_LITE
+    fprintf(stderr, "Running tests with regular db and operator.\n");
+    StringAppendOperatorTest::SetOpenDbFunction(&OpenNormalDb);
+  }
+
   typedef std::shared_ptr<DB> (* OpenFuncPtr)(char);
 
   // Allows user to open databases with different configurations.
@@ -129,7 +152,7 @@ StringAppendOperatorTest::OpenFuncPtr StringAppendOperatorTest::OpenDb = nullptr
 
 // THE TEST CASES BEGIN HERE
 
-TEST_F(StringAppendOperatorTest, IteratorTest) {
+TEST_P(StringAppendOperatorTest, IteratorTest) {
   auto db_ = OpenDb(',');
   StringLists slists(db_);
 
@@ -142,7 +165,8 @@ TEST_F(StringAppendOperatorTest, IteratorTest) {
   slists.Append("k2", "a3");
 
   std::string res;
-  std::unique_ptr<rocksdb::Iterator> it(db_->NewIterator(ReadOptions()));
+  std::unique_ptr<ROCKSDB_NAMESPACE::Iterator> it(
+      db_->NewIterator(ReadOptions()));
   std::string k1("k1");
   std::string k2("k2");
   bool first = true;
@@ -219,10 +243,9 @@ TEST_F(StringAppendOperatorTest, IteratorTest) {
       ASSERT_EQ(res, "g1");
     }
   }
-
 }
 
-TEST_F(StringAppendOperatorTest, SimpleTest) {
+TEST_P(StringAppendOperatorTest, SimpleTest) {
   auto db = OpenDb(',');
   StringLists slists(db);
 
@@ -237,7 +260,7 @@ TEST_F(StringAppendOperatorTest, SimpleTest) {
   ASSERT_EQ(res, "v1,v2,v3");
 }
 
-TEST_F(StringAppendOperatorTest, SimpleDelimiterTest) {
+TEST_P(StringAppendOperatorTest, SimpleDelimiterTest) {
   auto db = OpenDb('|');
   StringLists slists(db);
 
@@ -250,7 +273,7 @@ TEST_F(StringAppendOperatorTest, SimpleDelimiterTest) {
   ASSERT_EQ(res, "v1|v2|v3");
 }
 
-TEST_F(StringAppendOperatorTest, OneValueNoDelimiterTest) {
+TEST_P(StringAppendOperatorTest, OneValueNoDelimiterTest) {
   auto db = OpenDb('!');
   StringLists slists(db);
 
@@ -261,7 +284,7 @@ TEST_F(StringAppendOperatorTest, OneValueNoDelimiterTest) {
   ASSERT_EQ(res, "single_val");
 }
 
-TEST_F(StringAppendOperatorTest, VariousKeys) {
+TEST_P(StringAppendOperatorTest, VariousKeys) {
   auto db = OpenDb('\n');
   StringLists slists(db);
 
@@ -287,7 +310,7 @@ TEST_F(StringAppendOperatorTest, VariousKeys) {
 }
 
 // Generate semi random keys/words from a small distribution.
-TEST_F(StringAppendOperatorTest, RandomMixGetAppend) {
+TEST_P(StringAppendOperatorTest, RandomMixGetAppend) {
   auto db = OpenDb(' ');
   StringLists slists(db);
 
@@ -335,10 +358,9 @@ TEST_F(StringAppendOperatorTest, RandomMixGetAppend) {
     }
 
   }
-
 }
 
-TEST_F(StringAppendOperatorTest, BIGRandomMixGetAppend) {
+TEST_P(StringAppendOperatorTest, BIGRandomMixGetAppend) {
   auto db = OpenDb(' ');
   StringLists slists(db);
 
@@ -386,10 +408,9 @@ TEST_F(StringAppendOperatorTest, BIGRandomMixGetAppend) {
     }
 
   }
-
 }
 
-TEST_F(StringAppendOperatorTest, PersistentVariousKeys) {
+TEST_P(StringAppendOperatorTest, PersistentVariousKeys) {
   // Perform the following operations in limited scope
   {
     auto db = OpenDb('\n');
@@ -456,7 +477,7 @@ TEST_F(StringAppendOperatorTest, PersistentVariousKeys) {
   }
 }
 
-TEST_F(StringAppendOperatorTest, PersistentFlushAndCompaction) {
+TEST_P(StringAppendOperatorTest, PersistentFlushAndCompaction) {
   // Perform the following operations in limited scope
   {
     auto db = OpenDb('\n');
@@ -466,7 +487,7 @@ TEST_F(StringAppendOperatorTest, PersistentFlushAndCompaction) {
 
     // Append, Flush, Get
     slists.Append("c", "asdasd");
-    db->Flush(rocksdb::FlushOptions());
+    db->Flush(ROCKSDB_NAMESPACE::FlushOptions());
     success = slists.Get("c", &c);
     ASSERT_TRUE(success);
     ASSERT_EQ(c, "asdasd");
@@ -474,7 +495,7 @@ TEST_F(StringAppendOperatorTest, PersistentFlushAndCompaction) {
     // Append, Flush, Append, Get
     slists.Append("a", "x");
     slists.Append("b", "y");
-    db->Flush(rocksdb::FlushOptions());
+    db->Flush(ROCKSDB_NAMESPACE::FlushOptions());
     slists.Append("a", "t");
     slists.Append("a", "r");
     slists.Append("b", "2");
@@ -545,14 +566,14 @@ TEST_F(StringAppendOperatorTest, PersistentFlushAndCompaction) {
 
     // Append, Flush, Compact, Get
     slists.Append("b", "afcg");
-    db->Flush(rocksdb::FlushOptions());
+    db->Flush(ROCKSDB_NAMESPACE::FlushOptions());
     db->CompactRange(CompactRangeOptions(), nullptr, nullptr);
     slists.Get("b", &b);
     ASSERT_EQ(b, "y\n2\nmonkey\ndf\nl;\nafcg");
   }
 }
 
-TEST_F(StringAppendOperatorTest, SimpleTestNullDelimiter) {
+TEST_P(StringAppendOperatorTest, SimpleTestNullDelimiter) {
   auto db = OpenDb('\0');
   StringLists slists(db);
 
@@ -575,26 +596,13 @@ TEST_F(StringAppendOperatorTest, SimpleTestNullDelimiter) {
   ASSERT_EQ(res, checker);
 }
 
-} // namespace rocksdb
+INSTANTIATE_TEST_CASE_P(StringAppendOperatorTest, StringAppendOperatorTest,
+                        testing::Bool());
+
+}  // namespace ROCKSDB_NAMESPACE
 
 int main(int argc, char** argv) {
+  ROCKSDB_NAMESPACE::port::InstallStackTraceHandler();
   ::testing::InitGoogleTest(&argc, argv);
-  // Run with regular database
-  int result;
-  {
-    fprintf(stderr, "Running tests with regular db and operator.\n");
-    StringAppendOperatorTest::SetOpenDbFunction(&OpenNormalDb);
-    result = RUN_ALL_TESTS();
-  }
-
-#ifndef ROCKSDB_LITE  // TtlDb is not supported in Lite
-  // Run with TTL
-  {
-    fprintf(stderr, "Running tests with ttl db and generic operator.\n");
-    StringAppendOperatorTest::SetOpenDbFunction(&OpenTtlDb);
-    result |= RUN_ALL_TESTS();
-  }
-#endif  // !ROCKSDB_LITE
-
-  return result;
+  return RUN_ALL_TESTS();
 }
